import time
from contextlib import contextmanager


class TimerCount:
    def __init__(self, name: str, timer: "Timer"):
        self.name = name
        self.timer = timer
        self.start = time.time()

    def stop(self):
        self.timer.categories[self.name] += time.time() - self.start
        return self.timer


class Timer:
    instance = None

    def __init__(self):
        self.categories = {}

    @staticmethod
    def get_instance():
        if Timer.instance is None:
            Timer.instance = Timer()
        return Timer.instance

    @contextmanager
    def watch(self, name: str = "default"):
        try:
            count = self.start(name)
            yield None
        finally:
            count.stop()

    @contextmanager
    def watch_and_report(self, msg: str, print_fn=None):
        print_fn = print_fn or print
        try:
            start = time.time()
            yield None
        finally:
            end = time.time()
            print_fn(msg + f": {end - start:.3f} seconds")

    def start(self, name: str = "default"):
        if name not in self.categories:
            self.categories[name] = 0.0
        return TimerCount(name, self)

    def report(self, print_fn=None):
        print_fn = print_fn or print
        if len(self.categories) == 0:
            print_fn("--- Nothing to report ---")
            return

        print_fn("Runtime report:")
        for k, v in self.categories.items():
            print_fn(f"\t{k}: {v:.3f} seconds")

    def get_time(self, name: str = "default"):
        return self.categories[name]



import copy
import enum
import tempfile
from dataclasses import dataclass, asdict
from pathlib import Path
from typing import Optional, Union, Iterable, List, Set

import matplotlib.pyplot as plt
import networkx as nx
import orjson
import pydot
from IPython import get_ipython
from IPython.core.display import display
from PIL import Image

from sm.evaluation import sm_metrics
from sm.misc import auto_wrap


@dataclass
class SemanticType:
    class_abs_uri: str
    predicate_abs_uri: str
    class_rel_uri: str
    predicate_rel_uri: str

    @property
    def label(self):
        return (self.class_rel_uri, self.predicate_rel_uri)

    def is_entity_type(self) -> bool:
        """Telling if this semantic type is for entity column
        """
        return self.predicate_abs_uri in {
            'http://www.w3.org/2000/01/rdf-schema#label',
            'http://schema.org/name'
        }

    def __hash__(self):
        return hash((self.class_abs_uri, self.predicate_abs_uri))

    def __eq__(self, other):
        if not isinstance(other, SemanticType):
            return False

        return self.class_abs_uri == other.class_abs_uri and self.predicate_abs_uri == other.predicate_abs_uri

    def __str__(self):
        return f"{self.class_rel_uri}--{self.predicate_rel_uri}"

    def __repr__(self):
        return f"SType({self})"


@dataclass
class ClassNode:
    id: str
    abs_uri: str
    rel_uri: str
    approximation: bool = False
    readable_label: Optional[str] = None

    @property
    def label(self):
        return self.readable_label or self.rel_uri

    @property
    def is_class_node(self):
        return True

    @property
    def is_data_node(self):
        return False

    @property
    def is_literal_node(self):
        return False


@dataclass
class DataNode:
    id: str
    col_index: int
    label: str

    @property
    def is_class_node(self):
        return False

    @property
    def is_data_node(self):
        return True

    @property
    def is_literal_node(self):
        return False


class LiteralNodeDataType(str, enum.Enum):
    String = "string"
    Entity = "entity-id"


@dataclass
class LiteralNode:
    id: str
    value: str
    # readable label of the literal node! should not confuse it with value
    readable_label: Optional[str] = None
    # whether the literal node is in the surround context of the dataset
    is_in_context: bool = False
    datatype: LiteralNodeDataType = LiteralNodeDataType.String

    @property
    def label(self):
        return self.readable_label or self.value

    @property
    def is_class_node(self):
        return False

    @property
    def is_data_node(self):
        return False

    @property
    def is_literal_node(self):
        return True


Node = Union[ClassNode, DataNode, LiteralNode]


@dataclass
class Edge:
    source: str
    target: str
    abs_uri: str
    rel_uri: str
    approximation: bool = False
    readable_label: Optional[str] = None
    # id of the edge, this is set automatically by the semantic model in case we have multiple edges between nodes
    # and we need to delete one of them
    id: Optional[str] = None

    @property
    def label(self):
        return self.readable_label or self.rel_uri


class SemanticModel:
    def __init__(self, graph: Optional[nx.Graph] = None, edge_id_counter: int = 0):
        if graph is not None:
            self.g = graph
        else:
            self.g = nx.MultiDiGraph()
        self.edge_id_counter = edge_id_counter

    def get_n_nodes(self):
        return len(self.g.nodes)

    def get_n_edges(self):
        return len(self.g.edges)
    
    def in_degree(self, nid: str) -> int:
        return self.g.in_degree(nid)
    
    def out_degree(self, nid: str) -> int:
        return self.g.out_degree(nid)

    def degree(self, nid: str) -> int:
        return self.g.degree(nid)

    def get_node(self, nid: str) -> Optional[Node]:
        if nid not in self.g.nodes:
            return None
        return self.g.nodes[nid]['data']

    def get_data_node(self, column_index: int) -> Optional[DataNode]:
        for uid, u in self.g.nodes.data("data"):
            if not u.is_class_node and u.col_index == column_index:
                return u
        return None

    def get_literal_node(self, value: str) -> Optional[LiteralNode]:
        for uid, u in self.g.nodes.data("data"):
            if not u.is_literal_node and u.value == value:
                return u
        return None

    def get_edges_between_nodes(self, source_id: str, target_id: str) -> List[Edge]:
        res = self.g.get_edge_data(source_id, target_id)
        if res is None:
            return []
        return [x['data'] for x in res.values()]

    def has_node(self, nid: str):
        return self.g.has_node(nid)

    def has_edge(self, edge: Edge):
        return self.g.has_edge(edge.source, edge.target)

    def add_node(self, node: Node):
        self.g.add_node(node.id, data=node)

    def add_edge(self, edge: Edge):
        self.edge_id_counter += 1
        edge.id = self.edge_id_counter
        self.g.add_edge(edge.source, edge.target, key=self.edge_id_counter, data=edge)

    def update_node(self, nid: str, node: Node):
        """Update the node content by id"""
        if nid == node.id:
            # just update the node content
            self.g.nodes[nid]['data'] = node
        else:
            # has to replace all the edges
            inedges = list(self.g.in_edges(nid, data=True, keys=True))
            outedges = list(self.g.out_edges(nid, data=True, keys=True))

            self.g.add_node(node.id, data=node)
            for uid, vid, eid, edata in inedges:
                e: Edge = edata['data']
                e.target = node.id
                self.g.remove_edge(uid, vid, eid)
                self.g.add_edge(uid, node.id, eid, data=e)

            for uid, vid, eid, edata in outedges:
                e: Edge = edata['data']
                e.source = node.id
                self.g.remove_edge(uid, vid, eid)
                self.g.add_edge(node.id, vid, eid, data=e)

            self.g.remove_node(nid)

    def remove_node(self, node_id: str):
        self.g.remove_node(node_id)
    
    def remove_edge(self, edge: Edge):
        self.g.remove_edge(edge.source, edge.target, edge.id)

    def remove_edges_between_nodes(self, source_id: str, target_id: str, eid: Optional[str] = None):
        if eid is None:
            self.g.remove_edge(source_id, target_id)
        else:
            self.g.remove_edge(source_id, target_id, key=eid)

    def clone(self):
        return SemanticModel(copy.deepcopy(self.g), self.edge_id_counter)

    def iter_nodes(self) -> Iterable[Node]:
        return (u for uid, u in self.g.nodes.data("data"))

    def iter_edges(self) -> Iterable[Edge]:
        return (e for source, target, e in self.g.edges.data("data"))

    def incoming_edges(self, node_id: str) -> List[Edge]:
        """Get a list of incoming edges of a column"""
        lst = []
        for u, v, e in self.g.in_edges(node_id, data='data'):
            lst.append(e)
        return lst

    def outgoing_edges(self, node_id: str) -> List[Edge]:
        """Get a list of outgoing edges"""
        lst = []
        for u, v, e in self.g.out_edges(node_id, data='data'):
            lst.append(e)
        return lst

    def children(self, node_id: str) -> List[Node]:
        lst = []
        for u, v, e in self.g.out_edges(node_id, data='data'):
            lst.append(self.g.nodes[v]['data'])
        return lst

    def get_semantic_types_of_column(self, col_index: int) -> List[SemanticType]:
        dnode = self.get_data_node(col_index)
        sem_types = set()
        for u, v, e in self.g.in_edges(dnode.id, data='data'):
            u = self.get_node(u)
            sem_types.add(SemanticType(u.abs_uri, e.abs_uri, u.rel_uri, e.rel_uri))
        return list(sem_types)

    def get_semantic_types(self) -> Set[SemanticType]:
        sem_types = set()
        for e in self.iter_edges():
            u = self.get_node(e.source)
            assert isinstance(u, ClassNode)
            if self.get_node(e.target).is_class_node:
                continue

            sem_types.add(SemanticType(u.abs_uri, e.abs_uri, u.rel_uri, e.rel_uri))
        return sem_types

    def to_dict(self):
        return {
            "version": 1,
            "nodes": [asdict(u) for u in self.iter_nodes()],
            "edges": [asdict(e) for e in self.iter_edges()]
        }

    def to_json_file(self, outfile: Union[str, Path]):
        with open(outfile, "wb") as f:
            f.write(orjson.dumps(self.to_dict(), option=orjson.OPT_INDENT_2))

    @staticmethod
    def from_dict(record: dict):
        sm = SemanticModel()
        for u in record['nodes']:
            if 'col_index' in u:
                sm.add_node(DataNode(**u))
            elif 'abs_uri' in u:
                sm.add_node(ClassNode(**u))
            else:
                lnode = LiteralNode(**u)
                lnode.datatype = LiteralNodeDataType(lnode.datatype)
                sm.add_node(lnode)
        for e in record['edges']:
            assert sm.has_node(e['source']) and sm.has_node(e['target'])
            sm.add_edge(Edge(**e))
        return sm

    @staticmethod
    def from_json_file(infile: Union[str, Path]):
        with open(infile, "rb") as f:
            record = orjson.loads(f.read())
            return SemanticModel.from_dict(record)

    def draw(self, filename=None, no_display: bool = False, max_char_per_line: int = 20):
        """
        Parameters
        ----------
        filename : str | none
            output to a file or display immediately (inline if this is jupyter lab)

        no_display: bool
            if the code is running inside Jupyter, if enable, it returns the object and manually display (default is
            automatically display)

        max_char_per_line: int
            wrap the text if it's too long

        Returns
        -------
        """
        if filename is None:
            fobj = tempfile.NamedTemporaryFile()
            filename = fobj.name
        else:
            fobj = None

        dot_g = pydot.Dot(graph_type='digraph')
        for uid, u in self.g.nodes.data("data"):
            uid = uid.replace(":", "_")
            if u.is_class_node:
                label = auto_wrap(u.label.replace(":", "\:"), max_char_per_line)
                dot_g.add_node(
                    pydot.Node(name=uid, label=label, shape="ellipse", style="filled",
                               color="white",
                               fillcolor="lightgray"))
            elif u.is_data_node:
                label = auto_wrap(f"C{u.col_index}\:" + u.label.replace(":", "\:"), max_char_per_line)
                dot_g.add_node(
                    pydot.Node(name=uid, label=label, shape="plaintext", style="filled",
                               fillcolor="gold"))
            else:
                label = auto_wrap(u.value, max_char_per_line)
                dot_g.add_node(
                    pydot.Node(name=uid, label=label, shape="plaintext", style="filled",
                               fillcolor="purple"))

        for u, v, e in self.g.edges.data("data"):
            u = u.replace(":", "_")
            v = v.replace(":", "_")
            label = auto_wrap(e.label.replace(":", "\:"), max_char_per_line)
            dot_g.add_edge(pydot.Edge(u, v, label=label, color="brown", fontcolor="black"))

        # graphviz from anaconda does not support jpeg so use png instead
        dot_g.write(filename, prog='dot', format='png')

        if fobj is not None:
            img = Image.open(filename)
            try:
                if no_display:
                    return img
            finally:
                fobj.close()

            try:
                shell = get_ipython().__class__.__name__
                if shell == 'ZMQInteractiveShell':
                    display(img)
                else:
                    plt.imshow(img, interpolation='antialiased')
                    plt.show()
            except NameError:
                plt.imshow(img, interpolation='antialiased')
                plt.show()
            finally:
                fobj.close()

    def draw_difference(self, gold_sm: 'SemanticModel', filename=None, no_display: bool = False,
                        max_char_per_line: int = 20):
        """
        Colors:
        * green, red for edges/nodes in the pred_sm that does not appear in the gold_sm
        * lightgray for edges/nodes that are in the gold_sm but not in the pred_sm

        Parameters
        ----------
        gold_sm : SemanticModel
            the correct semantic model that we are going to compare to
        filename : str | none
            output to a file or display immediately (inline if this is jupyter lab)

        no_display : bool
            if the code is running inside Jupyter, if enable, it returns the object and manually display (default is
            automatically display)

        max_char_per_line: int
            wrap the text if it's too long

        Returns
        -------
        """
        if filename is None:
            fobj = tempfile.NamedTemporaryFile()
            filename = fobj.name
        else:
            fobj = None

        bijection = sm_metrics.precision_recall_f1(gold_sm, self)['_bijection']
        dot_g = pydot.Dot(graph_type='digraph')
        data_nodes = set()
        for uid, u in self.g.nodes.data("data"):
            if u.is_class_node:
                if bijection.prime2x[uid] is None:
                    # this is a wrong node
                    fillcolor = 'tomato'
                else:
                    fillcolor = 'mediumseagreen'

                label = auto_wrap(u.label.replace(":", "\:"), max_char_per_line)
                dot_g.add_node(
                    pydot.Node(name=uid.replace(":", "_"), label=label, shape="ellipse",
                               style="filled",
                               color="white",
                               fillcolor=fillcolor))
            else:
                data_nodes.add(u.col_index)
                dot_uid = f"C{u.col_index:02d}_{u.label}"
                label = auto_wrap(f"C{u.col_index}: " + u.label.replace(":", "\:"), max_char_per_line)
                dot_g.add_node(
                    pydot.Node(name=dot_uid, label=label, shape="plaintext", style="filled",
                               fillcolor="gold"))

        # node in gold_sm doesn't appear in the pred_sm
        for uid, u in gold_sm.g.nodes.data("data"):
            if u.is_class_node:
                if bijection.x2prime[uid] is None:
                    # class node in gold model need to give a different namespace (`gold:`) to avoid collision
                    dot_uid = ("gold:" + uid).replace(":", "_")
                    dot_g.add_node(
                        pydot.Node(name=dot_uid, label=auto_wrap(u.label.replace(":", "\:"), max_char_per_line),
                                   shape="ellipse",
                                   style="filled",
                                   color="white",
                                   fillcolor='lightgray'))
            else:
                if u.col_index not in data_nodes:
                    dot_uid = f"C{u.col_index:02d}_{u.label}"
                    dot_g.add_node(
                        pydot.Node(name=dot_uid,
                                   label=auto_wrap(f"C{u.col_index}: " + u.label.replace(":", "\:"), max_char_per_line),
                                   shape="plaintext",
                                   style="filled",
                                   fillcolor="lightgray"))

        # add edges in pred_sm
        x_triples = {
            (uid, e.label, vid if gold_sm.get_node(vid).is_class_node else (
            gold_sm.get_node(vid).col_index, gold_sm.get_node(vid).label))
            for uid, vid, e in gold_sm.g.edges.data("data")
        }
        x_prime_triples = set()
        for uid, vid, e in self.g.edges.data("data"):
            v = self.get_node(vid)
            x_prime_triple = (
            bijection.prime2x[uid], e.label, bijection.prime2x[vid] if v.is_class_node else (v.col_index, v.label))
            x_prime_triples.add(x_prime_triple)
            if x_prime_triple in x_triples:
                color = 'darkgreen'
            else:
                color = 'red'

            dot_u = uid.replace(":", "_")
            dot_v = vid.replace(":", "_") if v.is_class_node else f"C{v.col_index:02d}_{v.label}"
            dot_g.add_edge(pydot.Edge(dot_u, dot_v,
                                      label=auto_wrap(e.label.replace(":", "\:"), max_char_per_line),
                                      color=color, fontcolor="black"))

        # add edges in gold_sm that is not in pred_sm
        for x_triple in x_triples:
            if x_triple not in x_prime_triples:
                # class node in gold model need to give a different namespace (`gold:`) to avoid collision
                dot_u = "gold:" + x_triple[0] if bijection.x2prime[x_triple[0]] is None else bijection.x2prime[
                    x_triple[0]]
                dot_u = dot_u.replace(":", "_")

                if isinstance(x_triple[2], tuple):
                    dot_v = f"C{x_triple[2][0]:02d}_{x_triple[2][1]}"
                else:
                    dot_v = "gold:" + x_triple[2] if bijection.x2prime[x_triple[2]] is None else bijection.x2prime[
                        x_triple[2]]
                    dot_v = dot_v.replace(":", "_")

                dot_g.add_edge(pydot.Edge(dot_u, dot_v,
                                          label=auto_wrap(x_triple[1].replace(":", "\:"), max_char_per_line),
                                          color='gray', fontcolor="black"))

        # graphviz from anaconda does not support jpeg so use png instead
        dot_g.write(filename, prog='dot', format='png')

        if fobj is not None:
            img = Image.open(filename)
            try:
                if no_display:
                    return img
            finally:
                fobj.close()

            try:
                shell = get_ipython().__class__.__name__
                if shell == 'ZMQInteractiveShell':
                    display(img)
                else:
                    plt.imshow(img, interpolation='antialiased')
                    plt.show()
            except NameError:
                plt.imshow(img, interpolation='antialiased')
                plt.show()
            finally:
                fobj.close()



import pickle
from pathlib import Path
from typing import Dict, Union, Any, Callable

import orjson
import rocksdb

from sm.misc.funcs import identity_func
from sm.misc.big_dict.inmem import K, V, CacheDictStore


class RocksDBStore(Dict[K, V]):
    def __init__(self, dbfile: Union[Path, str], deserialize: Callable[[bytes], V] = None, create_if_missing=True, read_only=False):
        self.db = rocksdb.DB(str(dbfile), rocksdb.Options(create_if_missing=create_if_missing), read_only=read_only)
        if deserialize is not None:
            self.deserialize = deserialize
        else:
            self.deserialize = identity_func

    def __contains__(self, key):
        return self.db.get(key.encode()) is not None

    def __getitem__(self, key):
        item = self.db.get(key.encode())
        if item is None:
            raise KeyError(key)
        return self.deserialize(item)

    def __setitem__(self, key, value):
        self.db.put(key.encode(), value.encode())

    def __delitem__(self, key):
        self.db.delete(key.encode())

    def __len__(self):
        assert False, "Does not support this operator"

    def get(self, key: str, default=None):
        item = self.db.get(key.encode())
        if item is None:
            return None
        return self.deserialize(item)

    def cache_dict(self) -> 'CacheDictStore[K, V]':
        return CacheDictStore(self)


class JSONRocksDBStore(RocksDBStore[str, dict]):
    def __init__(self, dbfile: Union[Path, str], create_if_missing=True,
                 read_only=False):
        super().__init__(dbfile, deserialize=orjson.loads, create_if_missing=create_if_missing, read_only=read_only)

    def __setitem__(self, key, value):
        self.db.put(key.encode(), orjson.dumps(value))


class PickleRocksDBStore(RocksDBStore[K, V]):
    def __init__(self, dbfile: Union[Path, str], create_if_missing=True,
                 read_only=False):
        super().__init__(dbfile, deserialize=pickle.loads, create_if_missing=create_if_missing, read_only=read_only)

    def __setitem__(self, key: str, value: Any):
        value = pickle.dumps(value)
        self.db.put(key.encode(), value)



import os
from operator import itemgetter
from pathlib import Path

from sm.misc.deser import deserialize_yml


class OntNS:
    """A helper class for converting absolute uri to relative uri
    """
    instance = None

    def __init__(self):
        default_infile = Path(__file__).absolute().parent.parent / "data/namespaces.yml"
        infile = os.environ.get('OntNSFile', default_infile)
        if infile is not None:
            assert os.path.exists(infile)
        self.prefix2uri = dict(deserialize_yml(infile))
        # uri and prefix sorted by the length of uri
        self.uri_and_prefix = sorted([(uri, prefix) for prefix, uri in self.prefix2uri.items()], key=itemgetter(0),
                                     reverse=True)

    @staticmethod
    def get_instance() -> 'OntNS':
        if OntNS.instance is None:
            OntNS.instance = OntNS()
        return OntNS.instance

    def get_abs_uri(self, rel_uri: str):
        prefix, name = rel_uri.split(":", 2)
        return self.prefix2uri[prefix] + name

    def get_rel_uri(self, abs_uri: str):
        # default strategy, can make it faster by parsing the url and filter by the domain
        for uri, prefix in self.uri_and_prefix:
            if abs_uri.startswith(uri):
                return f"{prefix}:{abs_uri.replace(uri, '')}"
        raise Exception(f"Cannot simply the uri `{abs_uri}` as its namespace is not defined")



from sm.misc.fn_cache import *
from sm.misc.deser import *
from sm.misc.funcs import *
from sm.misc.big_dict import *
from sm.misc.exp_manager import *
from sm.misc.ontology_namespace import *
import sm.misc.graph as G


class UnreachableError(Exception):
    pass


from dataclasses import dataclass
from pathlib import Path
from typing import List, TypeVar, Generic, Union, Callable

from sm.misc import deserialize_json, get_latest_path
from sm.outputs import SemanticModel

T = TypeVar("T")


@dataclass
class Example(Generic[T]):
    sms: List[SemanticModel]
    table: T


def load(data_dir: Union[str, Path], table_deser: Callable[[dict], T]) -> List[Example]:
    """Load dataset from a folder. Assuming the following structure:
    - descriptions: (containing semantic descriptions of tables)
        - <table_fs_id>
            - version.01.json
            - version.02.json
            - ...
        - ...
    - tables: (containing list of tables, the type of table depends on )
        - <table_fs_id>.json
        - ...

    Args:
        data_dir:
        table_deser: deserialize the table from dictionary

    Returns:

    """
    data_dir = Path(data_dir)
    examples = []
    for infile in (data_dir / "tables").iterdir():
        if infile.name.startswith("."):
            continue
        assert infile.name.endswith(".json")
        example_id = infile.stem

        table = table_deser(deserialize_json(infile))
        raw_sms = deserialize_json(
            get_latest_path(data_dir / f"descriptions/{example_id}/version.json")
        )
        sms = [SemanticModel.from_dict(sm) for sm in raw_sms]

        examples.append(Example(sms=sms, table=table))
    return examples



from collections import OrderedDict, defaultdict
from typing import List, Optional

import pandas as pd

from sm.inputs.column import Column


class ColumnBasedTable:
    def __init__(self, table_id: str, columns: List[Column]):
        self.table_id = table_id
        self.columns = columns
        self.index2columns = {col.index: col for col in columns}
        self.df = self.as_dataframe()

    def get_column_by_index(self, col_idx: int) -> Column:
        return self.index2columns[col_idx]

    def as_dataframe(self) -> pd.DataFrame:
        d = OrderedDict()
        dup = defaultdict(int)
        for i, col in enumerate(self.columns):
            if col.name is None:
                cname = f"unk_{i:02}"
            else:
                cname = col.name

            if cname in d:
                dup[cname] += 1
                cname += f" ({dup[cname]})"
            d[cname] = col.values
        return pd.DataFrame(d)

    def subset(self, start_row: int, end_row: int):
        return ColumnBasedTable(
            self.table_id,
            [Column(c.index, c.name, c.values[start_row:end_row]) for c in self.columns],
        )

    def to_dict(self):
        return {
            "version": "2",
            "table_id": self.table_id,
            "columns": [col.__dict__ for col in self.columns],
        }

    @staticmethod
    def from_dict(record: dict):
        assert record.get('version', None) == "2"
        return ColumnBasedTable(record['table_id'], [Column(**col) for col in record['columns']])

    @staticmethod
    def from_dataframe(df: pd.DataFrame, table_id: str):
        columns = []
        for ci, c in enumerate(df.columns):
            values = [r[ci] for ri, r in df.iterrows()]
            column = Column(ci, c, values)
            columns.append(column)
        return ColumnBasedTable(table_id, columns)



import sys


# --------------------------------------------------------------------------------- #

def _input(filename):
    prices = {}
    names = {}

    for line in open(filename).readlines():
        (name, src, dst, price) = line.rstrip().split()
        name = int(name.replace('M', ''))
        src = int(src.replace('C', ''))
        dst = int(dst.replace('C', ''))
        price = int(price)
        t = (src, dst)
        if t in prices and prices[t] <= price:
            continue
        prices[t] = price
        names[t] = name

    return prices, names


def _load(arcs, weights):
    g = {}
    for (src, dst) in arcs:
        if src in g:
            g[src][dst] = weights[(src, dst)]
        else:
            g[src] = {dst: weights[(src, dst)]}
    return g


def _reverse(graph):
    r = {}
    for src in graph:
        for (dst, c) in graph[src].items():
            if dst in r:
                r[dst][src] = c
            else:
                r[dst] = {src: c}
    return r


def _getCycle(n, g, visited=None, cycle=None):
    if visited is None:
        visited = set()
    if cycle is None:
        cycle = []
    visited.add(n)
    cycle += [n]
    if n not in g:
        return cycle
    for e in g[n]:
        if e not in visited:
            cycle = _getCycle(e, g, visited, cycle)
    return cycle


def _mergeCycles(cycle, G, RG, g, rg):
    allInEdges = []
    minInternal = None
    minInternalWeight = float('inf')

    # find minimal internal edge weight
    for n in cycle:
        for e in RG[n]:
            if e in cycle:
                if minInternal is None or RG[n][e] < minInternalWeight:
                    minInternal = (n, e)
                    minInternalWeight = RG[n][e]
                    continue
            else:
                allInEdges.append((n, e))

                # find the incoming edge with minimum modified cost
    minExternal = None
    minModifiedWeight = 0
    for s, t in allInEdges:
        u, v = rg[s].popitem()
        rg[s][u] = v
        w = RG[s][t] - (v - minInternalWeight)
        if minExternal is None or minModifiedWeight > w:
            minExternal = (s, t)
            minModifiedWeight = w

    u, w = rg[minExternal[0]].popitem()
    rem = (minExternal[0], u)
    rg[minExternal[0]].clear()
    if minExternal[1] in rg:
        rg[minExternal[1]][minExternal[0]] = w
    else:
        rg[minExternal[1]] = {minExternal[0]: w}
    if rem[1] in g:
        if rem[0] in g[rem[1]]:
            del g[rem[1]][rem[0]]
    if minExternal[1] in g:
        g[minExternal[1]][minExternal[0]] = w
    else:
        g[minExternal[1]] = {minExternal[0]: w}


# --------------------------------------------------------------------------------- #

def mst(root, G):
    """ The Chu-Lui/Edmond's algorithm

    arguments:

    root - the root of the MST
    G - the graph in which the MST lies

    returns: a graph representation of the MST

    Graph representation is the same as the one found at:
    http://code.activestate.com/recipes/119466/

    Explanation is copied verbatim here:

    The input graph G is assumed to have the following
    representation: A vertex can be any object that can
    be used as an index into a dictionary.  G is a
    dictionary, indexed by vertices.  For any vertex v,
    G[v] is itself a dictionary, indexed by the neighbors
    of v.  For any edge v->w, G[v][w] is the length of
    the edge.  This is related to the representation in
    <http://www.python.org/doc/essays/graphs.html>
    where Guido van Rossum suggests representing graphs
    as dictionaries mapping vertices to lists of neighbors,
    however dictionaries of edges have many advantages
    over lists: they can store extra information (here,
    the lengths), they support fast existence tests,
    and they allow easy modification of the graph by edge
    insertion and removal.  Such modifications are not
    needed here but are important in other graph algorithms.
    Since dictionaries obey iterator protocol, a graph
    represented as described here could be handed without
    modification to an algorithm using Guido's representation.

    Of course, G and G[v] need not be Python dict objects;
    they can be any other object that obeys dict protocol,
    for instance a wrapper in which vertices are URLs
    and a call to G[v] loads the web page and finds its links.
    """

    RG = _reverse(G)
    if root in RG:
        RG[root] = {}
    g = {}
    for n in RG:
        if len(RG[n]) == 0:
            continue
        minimum = float("inf")
        s, d = None, None
        for e in RG[n]:
            if RG[n][e] < minimum:
                minimum = RG[n][e]
                s, d = n, e
        if d in g:
            g[d][s] = RG[s][d]
        else:
            g[d] = {s: RG[s][d]}

    cycles = []
    visited = set()
    for n in g:
        if n not in visited:
            cycle = _getCycle(n, g, visited)
            cycles.append(cycle)

    rg = _reverse(g)
    for cycle in cycles:
        if root in cycle:
            continue
        _mergeCycles(cycle, G, RG, g, rg)

    return g


# --------------------------------------------------------------------------------- #

if __name__ == "__main__":
    try:
        filename = sys.argv[1]
        root = sys.argv[2]
    except IndexError:
        sys.stderr.write('no input and/or root node specified\n')
        sys.stderr.write('usage: python edmonds.py <file> <root>\n')
        sys.exit(1)

    prices, names = _input(filename)
    g = _load(prices, prices)
    h = mst(int(root), g)
    for s in h:
        for t in h[s]:
            print("%d-%d" % (s, t))


import operator
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Union, List, Any, Optional

import networkx as nx


class PropPath:
    """A way to specify how to access property of a node/edge in networkx graph.
    Nested property is separated by a dot. For example: `data.column.name`.

    Id of a node or key of edge can be accessed using this path `@id`.
    """

    def __init__(self, path: str, allow_unknown_key: bool):
        self.origin_path = path
        self.allow_unknown_key = allow_unknown_key
        self.attrs = path.split(".")

        if self.attrs[0] == '@id':
            assert len(self.attrs) == 1

    def get_value(self, objectid: str, objectdata: dict):
        """Get value of the edge or node specified by the path

        Parameters
        ----------
        objectid: node id or edge key
        objectdata: node or edge data

        Returns
        -------

        """
        if self.attrs[0] == '@id':
            return objectid

        for attr in self.attrs:
            if isinstance(objectdata, dict):
                objectdata = objectdata[attr]
            elif hasattr(objectdata, attr):
                objectdata = getattr(objectdata, attr)
            else:
                if self.allow_unknown_key:
                    return None
                raise KeyError(attr)
        return objectdata

    def copy_to_dict(self, objectid: str, objectdata: dict, destination: dict):
        value = self.get_value(objectid, objectdata)
        assert not self.attrs[0] == '@id'

        for attr in self.attrs[:-1]:
            if attr not in destination:
                destination[attr] = {}
            destination = destination[attr]
        destination[self.attrs[-1]] = value


@dataclass
class Operator(Enum):
    Equal = "="
    LT = "<"
    LTE = "<="
    GT = ">"
    GTE = ">="


class Condition(ABC):
    @abstractmethod
    def satisfy(self, objectid: str, object: dict):
        """Whether the condition is satisfied"""
        pass


class TrueCondition(Condition):

    def satisfy(self, objectid: str, object: dict):
        return True


class BasicCondition(Condition):

    def __init__(self, path: PropPath, ops: Operator, target_value: Any):
        if ops == Operator.Equal:
            self.ops = operator.eq
        elif ops == Operator.LT:
            self.ops = operator.lt
        elif ops == Operator.LTE:
            self.ops = operator.le
        elif ops == Operator.GT:
            self.ops = operator.gt
        elif ops == Operator.GTE:
            self.ops = operator.ge
        else:
            raise Exception("Unreachable!")

        self.path = path
        self.target_value = target_value

    def satisfy(self, objectid: str, object: dict):
        value = self.path.get_value(objectid, object)
        return self.ops(value, self.target_value)


class NotCondition(Condition):

    def __init__(self, condition: 'Condition'):
        self.condition = condition

    def satisfy(self, objectid: str, object: dict):
        return not self.condition.satisfy(objectid, object)


class AndCondition(Condition):

    def __init__(self, conditions: List['Condition']):
        self.conditions = conditions

    def satisfy(self, objectid: str, object: dict):
        return all(condition.satisfy(objectid, object) for condition in self.conditions)


class OrCondition(Condition):

    def __init__(self, conditions: List['Condition']):
        self.conditions = conditions

    def satisfy(self, objectid: str, object: dict):
        return any(condition.satisfy(objectid, object) for condition in self.conditions)


@dataclass
class NodeVariable:
    # id of the node in the query
    id: str
    condition: Condition


@dataclass
class EdgeVariable:
    source_id: str
    target_id: str
    condition: Condition
    is_optional: bool = False


class QueryBuilder:
    def __init__(self):
        self.node_ids = set()
        self.variables = []

    def n(self, id: Optional[str] = None, condition: Optional['QueryConditionBuilder'] = None):
        """Return a node variable builder"""
        if id is None:
            id = str(uuid.uuid4())
        assert id not in self.node_ids, "Re-defined node id"
        if condition is None:
            condition = TrueCondition()
        else:
            condition = condition.build()
        node = NodeVariable(id, condition)
        self.variables.append(node)
        return QueryNodeBuilder(self, node)

    def c(self, path: str, ops: str, value: Any):
        """Return a condition"""
        return QueryConditionBuilder(self, BasicCondition(PropPath(path), Operator(ops), value))

    def get_query(self):
        return self.variables


@dataclass
class QueryConditionBuilder:
    builder: QueryBuilder
    condition: Condition

    def build(self) -> Condition:
        return self.condition

    def __and__(self, other: 'QueryConditionBuilder'):
        return QueryConditionBuilder(self.builder, AndCondition([self.condition, other.condition]))

    def __or__(self, other: 'QueryConditionBuilder'):
        return QueryConditionBuilder(self.builder, OrCondition([self.condition, other.condition]))

    def __neg__(self):
        return QueryConditionBuilder(self.builder, NotCondition(self.condition))


@dataclass
class QueryNodeBuilder:
    builder: QueryBuilder
    node_variable: NodeVariable

    def to(self, other: 'QueryNodeBuilder', condition: Optional['QueryConditionBuilder'] = None):
        return self._make_edge(other, condition, False)

    def to_optional(self, other: 'QueryNodeBuilder', condition: Optional['QueryConditionBuilder'] = None):
        return self._make_edge(other, condition, True)

    def _make_edge(self, other: 'QueryNodeBuilder', condition: Optional['QueryConditionBuilder'] = None, optional: bool = False):
        """Add an edge to the query"""
        if condition is None:
            condition = TrueCondition()
        else:
            condition = condition.build()
        edge = EdgeVariable(self.node_variable.id, other.node_variable.id, condition)
        self.builder.variables.append(edge)
        return other


def query_graph(g: nx.MultiDiGraph, queries: List[Union[NodeVariable, EdgeVariable]]):
    """
    Query the networkx graph to select a subset of the graph.

    The algorithm works like this:
    We first build a mapping, where each node variable is associated to
    the list of matched nodes in the graph.
    Then, we loop through each edge variable in the query, and record matched edges in the graph

    Then, we loop through each edge variable in the query that connects two node variables,
    for each edge in the graph, if it does not satisfied the query condition, we remove
    its the source and target from the corresponding node variables.
    We do a while loop until there is no more nodes getting removed
    Then, we do another final loop to find all the edges.
    """
    node_vars = {}
    for query in queries:
        if isinstance(query, NodeVariable):
            matched_nodes = set()
            for uid in g:
                u = g.nodes[uid]
                if query.condition.satisfy(uid, u):
                    matched_nodes.add(uid)
            node_vars[query.id] = matched_nodes
        elif isinstance(query, EdgeVariable):
            pass
        else:
            raise NotImplementedError()

    stop = False
    while not stop:
        stop = True
        for query in queries:
            if isinstance(query, NodeVariable):
                pass
            elif isinstance(query, EdgeVariable) and not query.is_optional:
                source_ids = node_vars[query.source_id]
                target_ids = node_vars[query.target_id]

                satisfied_source_ids = set()
                satisfied_target_ids = set()

                for uid, vid, eid, edata in g.edges(data=True, keys=True):
                    if uid not in source_ids:
                        continue
                    if vid not in target_ids:
                        continue
                    if query.condition.satisfy(eid, edata):
                        satisfied_source_ids.add(uid)
                        satisfied_target_ids.add(vid)

                deleted_sources = source_ids.difference(satisfied_source_ids)
                deleted_targets = target_ids.difference(satisfied_target_ids)
                if len(deleted_sources) + len(deleted_targets) > 0:
                    stop = False
                    for uid in deleted_sources:
                        source_ids.remove(uid)
                    for vid in deleted_targets:
                        target_ids.remove(vid)
            else:
                raise NotImplementedError()

    matched_edges = set()
    for query in queries:
        if isinstance(query, NodeVariable):
            pass
        elif isinstance(query, EdgeVariable):
            source_ids = node_vars[query.source_id]
            target_ids = node_vars[query.target_id]

            for uid, vid, eid, edata in g.edges(data=True, keys=True):
                if uid not in source_ids:
                    continue
                if vid not in target_ids:
                    continue
                if query.condition.satisfy(eid, edata):
                    matched_edges.add((uid, vid, eid))
        else:
            raise NotImplementedError()

    # now create a new graph
    resp_g = nx.MultiDiGraph()
    for ids in node_vars.values():
        for id in ids:
            if id not in resp_g:
                resp_g.add_node(id, **g.nodes[id])

    for uid, vid, eid in matched_edges:
        resp_g.add_edge(uid, vid, key=eid, **g.edges[uid, vid, eid])
    return resp_g



from dataclasses import dataclass
from typing import Set, Optional, Dict

from sm.evaluation import sm_metrics

from sm.outputs import SemanticModel, DataNode
from sm.evaluation.transformation import SemModelTransformation


@dataclass
class CTAEvalOutput:
    precision: float
    recall: float
    f1: float

    n_corrects: float  # float as we allow for partial correctness
    n_examples: int
    n_predictions: int


def cpa(
    gold_sm: SemanticModel,
    pred_sm: SemanticModel,
    id_props: Set[str],
    scoring_fn: Optional[sm_metrics.ScoringFn] = None,
) -> sm_metrics.PrecisionRecallF1Output:
    gold_sm = gold_sm.clone()
    pred_sm = pred_sm.clone()

    _cpa_transformation(gold_sm, id_props)
    _cpa_transformation(pred_sm, id_props)

    return sm_metrics.precision_recall_f1(
        gold_sm=gold_sm, pred_sm=pred_sm, scoring_fn=scoring_fn
    )


def cta(
    gold_sm: SemanticModel,
    pred_sm: SemanticModel,
    id_props: Set[str],
    scoring_fn: Optional[sm_metrics.ScoringFn] = None,
) -> CTAEvalOutput:
    gold_cta = _get_cta(gold_sm, id_props)
    pred_cta = _get_cta(pred_sm, id_props)

    if scoring_fn is None:
        scoring_fn = sm_metrics.ScoringFn()

    score = 0.0
    for cindex in set(gold_cta.keys()).intersection(pred_cta.keys()):
        gc = gold_cta[cindex]
        pc = pred_cta[cindex]
        score += scoring_fn.get_match_score(pc, gc)

    if len(pred_cta) == 0:
        precision = 1.0
    else:
        precision = score / len(pred_cta)

    if len(gold_cta) == 0:
        recall = 1.0
    else:
        recall = score / len(gold_cta)

    if recall + precision == 0:
        f1 = 0
    else:
        f1 = 2 * precision * recall / (precision + recall)
    return CTAEvalOutput(
        precision=precision,
        recall=recall,
        f1=f1,
        n_corrects=score,
        n_examples=len(gold_cta),
        n_predictions=len(pred_cta),
    )


def _get_cta(sm: SemanticModel, id_props: Set[str]) -> Dict[str, str]:
    col2class = {}
    for n in sm.iter_nodes():
        if n.is_class_node:
            outedges = sm.outgoing_edges(n.id)
            id_edges = [outedge for outedge in outedges if outedge.abs_uri in id_props]
            if len(id_edges) > 1:
                raise Exception("Haven't supported multiple subject columns yet")
            if len(id_edges) == 0:
                continue
            dnode: DataNode = sm.get_node(id_edges[0].target)
            col2class[str(dnode.col_index)] = n.abs_uri
    return col2class


def _cpa_transformation(sm: SemanticModel, id_props: Set[str]) -> None:
    SemModelTransformation.replace_class_nodes_by_subject_columns(sm, id_props)
    SemModelTransformation.remove_isolated_nodes(sm)



import functools

import orjson
from typing import Callable, Tuple, Any, Dict, Optional

from sm.misc.big_dict.rocksdb import PickleRocksDBStore
from sm.misc.big_dict.redis import PickleRedisStore

CACHE = {}


def cache_func(
    dbfile: str = "/tmp/cache_func.db",
    namespace: str = "",
    get_key: Callable[[str, str, Tuple[Any, ...], Dict[str, Any]], str] = None,
    instance_method: bool = False,
    cache: Optional[dict] = None,
):
    """Cache a function

    Args:
        dbfile: can be redis (e.g., "redis://localhost:6379") or rocksdb (local file)
        namespace:
        get_key:
        instance_method:
        cache:

    Returns:

    """
    global CACHE
    if cache is None:
        cache = CACHE

    if dbfile not in cache:
        if dbfile.startswith("redis://"):
            db = PickleRedisStore(dbfile)
        else:
            db = PickleRocksDBStore(dbfile)
        cache[dbfile] = db
    db = cache[dbfile]

    if get_key is None:
        get_key = default_get_key

    if instance_method:

        def wrapper_instance_fn(func):
            fn_name = func.__name__

            @functools.wraps(func)
            def fn(*args, **kwargs):
                key = get_key(namespace, fn_name, args[1:], kwargs)
                if key not in db:
                    db[key] = func(*args, **kwargs)
                return db[key]

            return fn

        return wrapper_instance_fn

    def wrapper_fn(func):
        fn_name = func.__name__

        @functools.wraps(func)
        def fn(*args, **kwargs):
            key = get_key(namespace, fn_name, args, kwargs)
            if key not in db:
                db[key] = func(*args, **kwargs)
            return db[key]

        return fn

    return wrapper_fn


def default_get_key(namespace, func_name, args, kwargs):
    return orjson.dumps(
        {"ns": namespace, "fn": func_name, "a": args, "kw": kwargs}
    ).decode()



import math
from typing import Dict, Callable, Set, Iterable

from sm.evaluation.sm_metrics import ScoringFn

ItemParents = Dict[str, Iterable[str]]


class HierarchyScoringFn(ScoringFn):
    def __init__(self, item_parents: ItemParents, get_item_uri: Callable[[str], str]):
        # mapping from the item to its parents
        self.i2i: Dict[str, Dict[str, int]] = {p: {} for p in item_parents}
        for p in item_parents:
            self.build_i2i(p, p, 1, item_parents, set())
        self.i2i = {
            get_item_uri(p): {get_item_uri(pp): d for pp, d in ppd.items()}
            for p, ppd in self.i2i.items()
        }

    def build_i2i(
        self,
        origin_item: str,
        current_item: str,
        distance: int,
        item_parents: ItemParents,
        visited: Set[str],
    ):
        if current_item in visited or current_item not in item_parents:
            return

        visited.add(current_item)
        for pp in item_parents[current_item]:
            if pp in self.i2i[origin_item]:
                self.i2i[origin_item][pp] = min(self.i2i[origin_item][pp], distance)
            else:
                self.i2i[origin_item][pp] = distance
            self.build_i2i(origin_item, pp, distance + 1, item_parents, visited)

    def get_match_score(self, pred_item: str, target_item: str):
        if pred_item == target_item:
            return 1.0
        if pred_item in self.i2i[target_item]:
            # pred_predicate is the parent of the target
            distance = self.i2i[target_item][pred_item]
            if distance > 5:
                return 0.0
            return math.pow(0.8, distance)
        if target_item in self.i2i[pred_item]:
            distance = self.i2i[pred_item][target_item]
            if distance > 3:
                return 0.0
            return math.pow(0.7, distance)
        return 0.0



import pickle
from typing import TypeVar, Dict, Union, Any

import redis

from sm.misc.big_dict.inmem import K, V, CacheDictStore


class RedisStore(Dict[K, V]):
    instance = None

    def __init__(self, url):
        self.url = url
        self.redis = redis.Redis.from_url(url)

    @classmethod
    def get_instance(cls):
        if cls.instance is None:
            cls.instance = cls()
        return cls.instance

    def __contains__(self, item: str):
        return self.redis.exists(item) == 1

    def __getitem__(self, item: str):
        assert item is not None, item
        resp = self.redis.get(item)
        if resp is None:
            raise KeyError(item)
        return self.deserialize(resp)

    def __setitem__(self, key: str, value: Union[str, bytes]):
        self.redis.set(key, value)

    def __len__(self):
        return self.redis.dbsize()

    def cache_dict(self) -> 'CacheDictStore[K, V]':
        """Return a version of the store that will cache all of the query for faster processing"""
        return CacheDictStore(self)

    def deserialize(self, value: str):
        return value


class PickleRedisStore(RedisStore[K, V]):

    def __setitem__(self, key: str, value: Any):
        value = pickle.dumps(value)
        self.redis.set(key, value)

    def deserialize(self, value: bytes):
        return pickle.loads(value)


import bz2
import csv
import gzip
import pickle

import chardet
import orjson
import json
from pathlib import Path
from typing import Union, List, Sequence, Optional, Any

import ujson
from ruamel.yaml import YAML


def get_open_fn(infile: Union[str, Path]):
    """Get the correct open function for the input file based on its extension. Supported bzip2, gz

    Parameters
    ----------
    infile : Union[str, Path]
        the file we wish to open

    Returns
    -------
    Callable
        the open function that can use to open the file

    Raises
    ------
    ValueError
        when encounter unknown extension
    """
    infile = str(infile)

    if infile.endswith(".bz2"):
        return bz2.open
    elif infile.endswith(".gz"):
        return gzip.open
    else:
        return open


def fix_encoding(fpath: Union[str, Path], backup_file: bool = True) -> bool:
    """Try to decode the context of the file as text in UTF-8, if it fails, try windows encoding before try to detect
    encoding of the file.

    If the encoding is not UTF-8,this function replace the content of the file with UTF-8 content and keep the backup
    if required.

    The function return True if the content is in UTF-8
    """
    with get_open_fn(str(fpath))(str(fpath), 'rb') as f:
        content = f.read()

    try:
        content = content.decode("utf-8")
        return True
    except UnicodeDecodeError:
        pass

    try:
        content = content.decode("windows-1252")
    except UnicodeDecodeError:
        encoding = chardet.detect(content)['encoding']
        content = content.decode(encoding)

    with get_open_fn(str(fpath))(str(fpath), 'wb') as f:
        f.write(content.encode())

    return content


def serialize_pkl(object, fpath: Union[str, Path]):
    """Serialize an object to a file"""
    with get_open_fn(str(fpath))(str(fpath), "wb") as f:
        pickle.dump(object, f)


def deserialize_pkl(fpath: Union[str, Path]):
    """Deserialize an object from a file"""
    with get_open_fn(str(fpath))(str(fpath), "rb") as f:
        return pickle.load(f)


def serialize_pkl_lines(objects: List[Any], fpath: Union[str, Path]):
    """Serialize a list of objects to a file"""
    with get_open_fn(str(fpath))(str(fpath), "wb") as f:
        for obj in objects:
            content = pickle.dumps(obj)
            f.write(len(content).to_bytes(4, 'little', signed=False))
            f.write(content)


def deserialize_pkl_lines(fpath: Union[str, Path]):
    """Deserialize a list of objects from a file"""
    with get_open_fn(str(fpath))(str(fpath), "rb") as f:
        objects = []
        while True:
            size = int.from_bytes(f.read(4), 'little', signed=False)
            if size == 0:
                break
            objects.append(pickle.loads(f.read(size)))
        return objects


def serialize_jl(objects, fpath: Union[str, Path]):
    """Serialize a list of objects to a file in JSON line format"""
    with get_open_fn(str(fpath))(str(fpath), "wb") as f:
        for obj in objects:
            f.write(orjson.dumps(obj, option=orjson.OPT_SERIALIZE_DATACLASS))
            f.write(b"\n")


def deserialize_jl(fpath: Union[str, Path], n_lines: Optional[int]=None):
    with get_open_fn(str(fpath))(str(fpath), "rb") as f:
        if n_lines is None:
            return [orjson.loads(line) for line in f]
        lst = []
        for line in f:
            lst.append(orjson.loads(line))
            if len(lst) >= n_lines:
                break
        return lst


def serialize_json(
    obj, fpath: Union[str, Path], use_pyjson: bool = False, indent: Optional[int] = None
):
    with get_open_fn(str(fpath))(str(fpath), "wb") as f:
        if use_pyjson:
            f.write(json.dumps(obj, indent=indent).encode())
        else:
            if indent is not None:
                f.write(
                    ujson.dumps(
                        obj, indent=indent, escape_forward_slashes=False
                    ).encode()
                )
            else:
                f.write(orjson.dumps(obj, option=orjson.OPT_SERIALIZE_DATACLASS))


def deserialize_json(fpath: Union[str, Path]):
    with get_open_fn(str(fpath))(str(fpath), "rb") as f:
        return orjson.loads(f.read())


def deserialize_byte_lines(fpath: Union[str, Path], n_lines: Optional[int]=None):
    """Deserialize byte lines, each line should never have byte b'\n'."""
    with get_open_fn(str(fpath))(str(fpath), "rb") as f:
        if n_lines is None:
            return [line for line in f]
        lst = []
        for line in f:
            lst.append(line)
            if len(lst) >= n_lines:
                break
        return lst


def serialize_byte_lines(objects: Sequence[bytes], fpath: Union[str, Path]):
    """Serialize byte lines, each line should never have byte b'\n'."""
    with get_open_fn(str(fpath))(str(fpath), "wb") as f:
        for obj in objects:
            f.write(obj)
            f.write(b"\n")


def serialize_lines(objects: Sequence[str], fpath: Union[str, Path]):
    with get_open_fn(str(fpath))(str(fpath), "wb") as f:
        for obj in objects:
            f.write(obj.encode())
            f.write(b"\n")


def deserialize_lines(fpath: Union[str, Path], n_lines: Optional[int]=None, trim: bool=False):
    with get_open_fn(str(fpath))(str(fpath), "rb") as f:
        if n_lines is None:
            if trim:
                return [line.decode().strip() for line in f]    
            return [line.decode() for line in f]

        lst = []
        if trim:
            for line in f:
                lst.append(line.decode().strip())
                if len(lst) >= n_lines:
                    break
        else:
            for line in f:
                lst.append(line.decode())
                if len(lst) >= n_lines:
                    break
        return lst


def deserialize_key_val_lines(fpath: Union[str, Path], delimiter="\t"):
    with get_open_fn(str(fpath))(str(fpath), "rb") as f:
        return [line.decode().split(delimiter, 1) for line in f]


def deserialize_key_json_val_lines(fpath: Union[str, Path], delimiter="\t"):
    with get_open_fn(str(fpath))(str(fpath), "rb") as f:
        results = []
        for line in f:
            k, v = line.decode().split(delimiter, 1)
            results.append((k, orjson.loads(v)))
        return results


def serialize_yml(objects, fpath: Union[str, Path]):
    with get_open_fn(fpath)(fpath, "wb") as f:
        yaml = YAML()
        yaml.dump(objects, f)


def deserialize_yml(fpath: Union[str, Path]):
    with get_open_fn(fpath)(fpath, "rb") as f:
        yaml = YAML()
        return yaml.load(f)


def serialize_csv(rows, fpath: Union[str, Path], mode="w", delimiter=","):
    with open(fpath, mode, newline="") as f:
        writer = csv.writer(
            f, delimiter=delimiter, quoting=csv.QUOTE_MINIMAL, lineterminator="\n"
        )
        for row in rows:
            writer.writerow(row)


def deserialize_csv(fpath: Union[str, Path], delimiter=","):
    with get_open_fn(fpath)(fpath, "rt") as f:
        reader = csv.reader(f, delimiter=delimiter)
        return [row for row in reader]


def serialize_text(text: str, fpath: Union[str, Path]):
    with get_open_fn(str(fpath))(str(fpath), "wb") as f:
        f.write(text.encode())


def serialize_bytes(text: bytes, fpath: Union[str, Path]):
    with get_open_fn(str(fpath))(str(fpath), "wb") as f:
        f.write(text)


def deserialize_text(fpath: Union[str, Path]):
    with get_open_fn(str(fpath))(str(fpath), "rb") as f:
        return f.read().decode()


def deserialize_bytes(fpath: Union[str, Path]):
    with get_open_fn(str(fpath))(str(fpath), "rb") as f:
        return f.read()



#!/usr/bin/python
# -*- coding: utf-8 -*-

import math
import os
from dataclasses import dataclass
from itertools import permutations, chain
from typing import Dict, Tuple, List, Set, Optional, Callable, Generator, TYPE_CHECKING

from loguru import logger
from pyrsistent import pvector, PVector

if TYPE_CHECKING:
    from sm.outputs import SemanticModel

"""
Compute precision, recall and f1 score of the semantic model according to Mohsen paper.

Convention: x' and x are nodes in predicted model and gold model, respectively.
"""


class PermutationExplosion(Exception):
    pass


class Node(object):
    def __init__(self, id: str, label: str) -> None:
        self.id: str = id
        self.label: str = label
        self.incoming_links: List[Link] = []
        self.outgoing_links: List[Link] = []

    @staticmethod
    def add_incoming_link(self: "Node", link: "Link"):
        self.incoming_links.append(link)
        link.target = self

    @staticmethod
    def add_outgoing_link(self: "Node", link: "Link"):
        self.outgoing_links.append(link)
        link.source = self

    def __str__(self):
        return "Node(id=%s, label=%s)" % (self.id, self.label)


@dataclass(frozen=True, eq=True)
class NodeTriple:
    __slots__ = ("source_id", "link_label", "target_id")
    source_id: str
    link_label: str
    target_id: str

    def __setstate__(self, state):
        assert state[0] is None and isinstance(state[1], dict)
        for slot, value in state[1].items():
            object.__setattr__(self, slot, value)


class Link(object):
    def __init__(self, id: int, label: str, source_id: str, target_id: str) -> None:
        self.id: int = id
        self.label: str = label
        self.target_id: str = target_id
        self.source_id: str = source_id
        # noinspection PyTypeChecker
        self.source: Node = None
        # noinspection PyTypeChecker
        self.target: Node = None


class LabelGroup(object):
    """Represent a group of nodes that have same label"""

    def __init__(self, nodes: List[Node]) -> None:
        self.nodes: List[Node] = nodes
        self.node_triples: Set[NodeTriple] = {
            NodeTriple(link.source_id, link.label, link.target_id)
            for node in self.nodes
            for link in chain(node.incoming_links, node.outgoing_links)
        }
        self.size: int = len(nodes)

    def __repr__(self):
        return "(#nodes=%d)" % (len(self.nodes))

    # noinspection PyUnusedLocal
    @staticmethod
    def group_by_structures(group: "LabelGroup", pred_group: "LabelGroup"):
        """
        A structure of a node is defined by its links, or we can treat it as a set of triple.
        Unbounded nodes should be assumed to be different, therefore a node have unbounded nodes will have
        it own structure group.
        We need not consider triple that are impossible to map to node in pred_group. This trick will improve
        the performance.
        """
        # TODO: implement it
        return [StructureGroup([n]) for n in group.nodes]


class StructureGroup(object):
    """Represent a group of nodes that have same structure"""

    def __init__(self, nodes: List[Node]) -> None:
        self.nodes: List[Node] = nodes
        self.size: int = len(nodes)


class PairLabelGroup(object):
    def __init__(self, label: str, X: LabelGroup, X_prime: LabelGroup) -> None:
        self.label = label
        self.X: LabelGroup = X
        self.X_prime: LabelGroup = X_prime

    def __repr__(self):
        return "(label=%s, X=%s, X_prime=%s)" % (self.label, self.X, self.X_prime)


class DependentGroups(object):
    """Represent a list of groups of nodes that are dependent on each other"""

    def __init__(self, pair_groups: List[PairLabelGroup]):
        self.pair_groups: List[PairLabelGroup] = pair_groups
        self.X_triples: Set[NodeTriple] = pair_groups[0].X.node_triples
        self.X_prime_triples: Set[NodeTriple] = pair_groups[0].X_prime.node_triples

        for pair in pair_groups[1:]:
            self.X_triples = self.X_triples.union(pair.X.node_triples)
            self.X_prime_triples = self.X_prime_triples.union(pair.X_prime.node_triples)

        # a mapping from (source id, target id) to list of predicates
        # use this for computing if we want to take into account the ancestor/descendant in the predicates
        self.x_pairs = {}
        for triple in self.X_triples:
            if (triple.source_id, triple.target_id) not in self.x_pairs:
                self.x_pairs[triple.source_id, triple.target_id] = []
            self.x_pairs[triple.source_id, triple.target_id].append(triple.link_label)

    def get_n_permutations(self):
        n_permutation = 1
        for pair_group in self.pair_groups:
            n = max(pair_group.X.size, pair_group.X_prime.size)
            m = min(pair_group.X.size, pair_group.X_prime.size)
            n_permutation *= math.factorial(n) / math.factorial(n - m)

        return n_permutation


class Bijection(object):
    """A bijection defined a one-one mapping from x' => x"""

    def __init__(self) -> None:
        # a map from x' => x (pred_sm to gold_sm)
        self.prime2x: Dict[str, str] = {}
        # map from x => x'
        self.x2prime: Dict[str, str] = {}

    @staticmethod
    def construct_from_mapping(
        mapping: List[Tuple[Optional[int], Optional[int]]]
    ) -> "Bijection":
        """
        :param mapping: a list of map from x' => x
        """
        self = Bijection()
        self.prime2x = {x_prime: x for x_prime, x in mapping}
        self.x2prime = {x: x_prime for x_prime, x in mapping}
        return self

    def extends(self, bijection: "Bijection") -> "Bijection":
        another = Bijection()
        another.prime2x = dict(self.prime2x)
        another.prime2x.update(bijection.prime2x)
        another.x2prime = dict(self.x2prime)
        another.x2prime.update(bijection.x2prime)
        return another

    def extends_(self, bijection: "Bijection") -> None:
        self.prime2x.update(bijection.prime2x)
        self.x2prime.update(bijection.x2prime)

    def is_gold_node_bounded(self, node_id: str) -> bool:
        return node_id in self.x2prime

    def is_pred_node_bounded(self, node_id: str) -> bool:
        return node_id in self.prime2x


class IterGroupMapsUsingGroupingArgs(object):
    def __init__(self, node_index: int, bijection: PVector, G_sizes) -> None:
        self.node_index: int = node_index
        self.bijection: PVector = bijection
        self.G_sizes = G_sizes


class IterGroupMapsGeneralApproachArgs(object):
    def __init__(self, node_index: int, bijection: PVector) -> None:
        self.node_index: int = node_index
        self.bijection: PVector = bijection


class FindBestMapArgs(object):
    def __init__(self, group_index: int, bijection: Bijection) -> None:
        self.group_index: int = group_index
        self.bijection = bijection


class ScoringFn:
    # noinspection PyMethodMayBeStatic
    def get_match_score(self, pred_predicate: str, target_predicate: str) -> float:
        return int(pred_predicate == target_predicate)


@dataclass
class PrecisionRecallF1Output:
    precision: float
    recall: float
    f1: float
    bijection: Bijection
    n_corrects: float  # float as we allow for partial correctness
    n_examples: int
    n_predictions: int
    gold_triples: Set[NodeTriple]
    pred_triples: Set[NodeTriple]


def find_best_map(
    dependent_group: DependentGroups, bijection: Bijection, scoring_fn: ScoringFn
) -> Bijection:
    terminate_index: int = len(dependent_group.pair_groups)
    # This code find the size of this
    # array: sum([min(gold_group.size, pred_group.size) for gold_group, pred_group in dependent_group.groups])
    call_stack = [FindBestMapArgs(group_index=0, bijection=bijection)]
    n_called = 0

    best_map = None
    best_score = -1

    while True:
        call_args = call_stack.pop()
        n_called += 1
        if call_args.group_index == terminate_index:
            # it is terminated, calculate score
            score = eval_score(dependent_group, call_args.bijection, scoring_fn)
            if score > best_score:
                best_score = score
                best_map = call_args.bijection
        else:
            pair_group = dependent_group.pair_groups[call_args.group_index]
            X, X_prime = pair_group.X, pair_group.X_prime
            for group_map in iter_group_maps(X, X_prime, call_args.bijection):
                bijection = Bijection.construct_from_mapping(group_map)

                call_stack.append(
                    FindBestMapArgs(
                        group_index=call_args.group_index + 1,
                        bijection=call_args.bijection.extends(bijection),
                    )
                )

        if len(call_stack) == 0:
            break

    return best_map


def get_unbounded_nodes(
    X: LabelGroup, is_bounded_func: Callable[[str], bool]
) -> List[Node]:
    """Get nodes of a label group which have not been bounded by a bijection"""
    unbounded_nodes = []

    for x in X.nodes:
        for link in x.incoming_links:
            if not is_bounded_func(link.source_id):
                unbounded_nodes.append(link.source)

        for link in x.outgoing_links:
            if not is_bounded_func(link.target_id):
                unbounded_nodes.append(link.target)

    return unbounded_nodes


def get_common_unbounded_nodes(
    X: LabelGroup, X_prime: LabelGroup, bijection: Bijection
) -> Set[str]:
    """Finding unbounded nodes in X and X_prime that have same labels"""
    unbounded_X = get_unbounded_nodes(X, bijection.is_gold_node_bounded)
    unbounded_X_prime = get_unbounded_nodes(X_prime, bijection.is_pred_node_bounded)

    labeled_unbounded_X = {}
    labeled_unbounded_X_prime = {}
    for x in unbounded_X:
        if x.label not in labeled_unbounded_X:
            labeled_unbounded_X[x.label] = []
        labeled_unbounded_X[x.label].append(x)

    for x in unbounded_X_prime:
        if x.label not in labeled_unbounded_X_prime:
            labeled_unbounded_X_prime[x.label] = []
        labeled_unbounded_X_prime[x.label].append(x)

    common_unbounded_nodes = set(labeled_unbounded_X.keys()).intersection(
        labeled_unbounded_X_prime.keys()
    )
    return common_unbounded_nodes


def group_dependent_elements(dependency_map: List[List[int]]) -> List[int]:
    # algorithm to merge the dependencies
    # input:
    #   - dependency_map: [<element_index, ...>, ...] list of dependencies where element at ith position is list of index of elements
    #           that element at ith position depends upon.
    # output:
    #   - dependency_groups: [<group_id>, ...] list of group id, where element at ith position is group id that element belongs to
    dependency_groups: List[int] = [-1 for _ in range(len(dependency_map))]
    invert_dependency_groups = {}

    for i, g in enumerate(dependency_map):
        dependent_elements = g + [i]
        groups = {dependency_groups[j] for j in dependent_elements}
        valid_groups = groups.difference([-1])
        if len(valid_groups) == 0:
            group_id = len(invert_dependency_groups)
            invert_dependency_groups[group_id] = set()
        else:
            group_id = next(iter(valid_groups))

        if -1 in groups:
            # map unbounded elements to group has group_id
            for j in dependent_elements:
                if dependency_groups[j] == -1:
                    dependency_groups[j] = group_id
                    invert_dependency_groups[group_id].add(j)

        for another_group_id in valid_groups.difference([group_id]):
            for j in invert_dependency_groups[another_group_id]:
                dependency_groups[j] = group_id
                invert_dependency_groups[group_id].add(j)

    return dependency_groups


def split_by_dependency(
    map_groups: List[PairLabelGroup], bijection: Bijection
) -> List[DependentGroups]:
    """This method takes a list of groups (X, X') and group them based on their dependencies.
    D = {D1, D2, …} s.t for all Di, Dj, (Xi, Xi') in Di, (Xj, Xj’) in Dj, they are independent

    Two groups of nodes are dependent when at least one unbounded nodes in a group is a label of other group.
    For example, "actor_appellation" has link to "type", so group "actor_appellation" depends on group "type"
    """
    group_label2idx = {map_group.label: i for i, map_group in enumerate(map_groups)}

    # build group dependencies
    dependency_map = [[] for _ in range(len(map_groups))]
    for i, map_group in enumerate(map_groups):
        X, X_prime = map_group.X, map_group.X_prime
        common_labels = get_common_unbounded_nodes(X, X_prime, bijection)

        for common_label in common_labels:
            group_id = group_label2idx[common_label]
            dependency_map[i].append(group_id)

    dependency_groups = group_dependent_elements(dependency_map)
    dependency_pair_groups: Dict[int, List[PairLabelGroup]] = {}
    dependency_map_groups: List[DependentGroups] = []

    for i, map_group in enumerate(map_groups):
        if dependency_groups[i] not in dependency_pair_groups:
            dependency_pair_groups[dependency_groups[i]] = []
        dependency_pair_groups[dependency_groups[i]].append(map_group)

    for pair_groups in dependency_pair_groups.values():
        dependency_map_groups.append(DependentGroups(pair_groups))

    return dependency_map_groups


# noinspection PyUnusedLocal
def iter_group_maps(
    X: LabelGroup, X_prime: LabelGroup, bijection: Bijection
) -> Generator[List[Tuple[int, int]], None, None]:
    if X.size < X_prime.size:
        return iter_group_maps_general_approach(X, X_prime)
    else:
        G = LabelGroup.group_by_structures(X, X_prime)
        return iter_group_maps_using_grouping(X_prime, G)


def iter_group_maps_general_approach(
    X: LabelGroup, X_prime: LabelGroup
) -> Generator[List[Tuple[int, int]], None, None]:
    """
    Generate all mapping from X to X_prime
    NOTE: |X| < |X_prime|

    Return mapping from (x_prime to x)
    """
    mapping_mold: List[Optional[str]] = [None for _ in range(X_prime.size)]

    for perm in permutations(range(X_prime.size), X.size):
        mapping: List[Tuple[str, str]] = []
        for i, j in enumerate(perm):
            mapping_mold[j] = X.nodes[i].id

        for i in range(X_prime.size):
            mapping.append((X_prime.nodes[i].id, mapping_mold[i]))
            mapping_mold[i] = None
        yield mapping


def iter_group_maps_using_grouping(
    X_prime: LabelGroup, G: List[StructureGroup]
) -> Generator[List[Tuple[int, int]], None, None]:
    """
    Generate all mapping from X_prime to G (nodes in X grouped by their structures)
    NOTE: |X_prime| <= |X|

    Return mapping from (x_prime to x)
    """
    G_sizes = pvector((g.size for g in G))
    bijection: PVector = pvector([-1 for _ in range(X_prime.size)])
    terminate_index: int = X_prime.size

    call_stack: List[IterGroupMapsUsingGroupingArgs] = [
        IterGroupMapsUsingGroupingArgs(
            node_index=0, bijection=bijection, G_sizes=G_sizes
        )
    ]

    while True:
        call_args = call_stack.pop()

        if call_args.node_index == terminate_index:
            # convert bijection into a final mapping
            G_numerator = [0 for _ in range(len(G))]
            bijection = call_args.bijection
            mapping = []
            for i in range(len(bijection)):
                x_prime = X_prime.nodes[i].id
                x = G[bijection[i]].nodes[G_numerator[bijection[i]]].id

                G_numerator[bijection[i]] += 1
                mapping.append((x_prime, x))

            yield mapping
        else:
            for i, G_i in enumerate(G):
                if call_args.G_sizes[i] == 0:
                    continue

                bijection = call_args.bijection.set(call_args.node_index, i)
                G_sizes = call_args.G_sizes.set(i, call_args.G_sizes[i] - 1)

                call_stack.append(
                    IterGroupMapsUsingGroupingArgs(
                        node_index=call_args.node_index + 1,
                        bijection=bijection,
                        G_sizes=G_sizes,
                    )
                )

        if len(call_stack) == 0:
            break


def prepare_args(
    gold_sm: "SemanticModel", pred_sm: "SemanticModel"
) -> List[PairLabelGroup]:
    """Prepare data for evaluation

    + data_node_mode = 0, mean we don't touch anything (note that the label of data_node must be unique)
    + data_node_mode = 1, mean we ignore label of data node (convert it to DATA_NODE, DATA_NODE2 if there are duplication columns)
    + data_node_mode = 2, mean we ignore data node
    """

    def convert_graph(graph: "SemanticModel"):
        node_index: Dict[str, Node] = {}

        for v in graph.iter_nodes():
            if v.is_class_node:
                label = v.abs_uri
            elif v.is_data_node:
                label = f"C{v.col_index:02d}:{v.label}"
            else:
                assert v.is_literal_node
                label = v.value
            node_index[v.id] = Node(v.id, label)

        for i, e in enumerate(graph.iter_edges()):
            link = Link(i, e.abs_uri, e.source, e.target)
            Node.add_outgoing_link(node_index[e.source], link)
            Node.add_incoming_link(node_index[e.target], link)

        return node_index

    label2nodes = {}
    gold_id2node = convert_graph(gold_sm)
    pred_id2node = convert_graph(pred_sm)

    for node in gold_id2node.values():
        if node.label not in label2nodes:
            label2nodes[node.label] = ([], [])
        label2nodes[node.label][0].append(node)
    for node in pred_id2node.values():
        if node.label not in label2nodes:
            label2nodes[node.label] = ([], [])
        label2nodes[node.label][1].append(node)

    return [
        PairLabelGroup(label, LabelGroup(g[0]), LabelGroup(g[1]))
        for label, g in label2nodes.items()
    ]


def eval_score(
    dependent_groups: DependentGroups, bijection: Bijection, scoring_fn: ScoringFn
) -> float:
    x_pairs = dependent_groups.x_pairs
    mapped_xprime_triples = {}
    for triple in dependent_groups.X_prime_triples:
        if (
            triple.source_id not in bijection.prime2x
            or triple.target_id not in bijection.prime2x
        ):
            continue

        s = bijection.prime2x[triple.source_id]
        o = bijection.prime2x[triple.target_id]
        if (s, o) in x_pairs:
            if (s, o) not in mapped_xprime_triples:
                mapped_xprime_triples[(s, o)] = []
            mapped_xprime_triples[(s, o)].append(triple.link_label)

    score = 0.0
    for so, edges in mapped_xprime_triples.items():
        if len(x_pairs[so]) == 1:
            gold_edge = x_pairs[so][0]
            # if save some computation time
            if len(edges) > 1:
                edge = max(
                    edges, key=lambda e: scoring_fn.get_match_score(e, gold_edge)
                )
            else:
                edge = edges[0]
            score += scoring_fn.get_match_score(edge, gold_edge)
        else:
            free_prime_edge_index = set(range(len(edges)))
            for gold_edge in x_pairs[so]:
                edge_idx = max(
                    free_prime_edge_index,
                    key=lambda idx: scoring_fn.get_match_score(edges[idx], gold_edge),
                )
                free_prime_edge_index.remove(edge_idx)
                score += scoring_fn.get_match_score(edges[edge_idx], gold_edge)

    return score


def precision_recall_f1(
    gold_sm: "SemanticModel",
    pred_sm: "SemanticModel",
    scoring_fn: Optional[ScoringFn] = None,
    debug_dir: str = None,
) -> PrecisionRecallF1Output:
    if scoring_fn is None:
        scoring_fn = ScoringFn()
    pair_groups: List[PairLabelGroup] = prepare_args(gold_sm, pred_sm)

    mapping = []
    map_groups: List[PairLabelGroup] = []
    for pair in pair_groups:
        X, X_prime = pair.X, pair.X_prime

        if max(X.size, X_prime.size) == 1:
            x_prime = None if X_prime.size == 0 else X_prime.nodes[0].id
            x = None if X.size == 0 else X.nodes[0].id
            mapping.append((x_prime, x))
        else:
            map_groups.append(pair)

    bijection = Bijection.construct_from_mapping(mapping)
    list_of_dependent_groups: List[DependentGroups] = split_by_dependency(
        map_groups, bijection
    )

    best_bijections = []
    n_permutations = sum(
        [
            dependent_groups.get_n_permutations()
            for dependent_groups in list_of_dependent_groups
        ]
    )

    # TODO: remove debugging code or change to logging
    if n_permutations > 50000:
        logger.info("Number of permutation is: {}", n_permutations)

    if n_permutations > 1000000:
        if debug_dir is not None:
            gold_sm.draw(os.path.join(debug_dir, "/gold.png"))
            pred_sm.draw(os.path.join(debug_dir, "/pred.png"))
        logger.error(
            "Permutation explosion: got {} combinations from {} pair groups",
            n_permutations,
            len(list_of_dependent_groups),
        )
        for dependent_groups in list_of_dependent_groups:
            logger.info("- {}", dependent_groups.pair_groups)
        raise PermutationExplosion(
            "Cannot run evaluation because number of permutation is too high."
        )

    for dependent_groups in list_of_dependent_groups:
        best_bijections.append(find_best_map(dependent_groups, bijection, scoring_fn))

    for best_bijection in best_bijections:
        bijection = bijection.extends(best_bijection)

    all_groups = DependentGroups(pair_groups)

    TP = eval_score(all_groups, bijection, scoring_fn)

    if len(all_groups.X_triples) == 0:
        # gold is empty, recall must be 1
        recall = 1
    else:
        recall = TP / len(all_groups.X_triples)

    if len(all_groups.X_prime_triples) == 0:
        # predict nothing, the precision must be 1
        precision = 1
    else:
        precision = TP / len(all_groups.X_prime_triples)

    if precision == 0 and recall == 0:
        f1 = 0.0
    else:
        f1 = 2 * precision * recall / (precision + recall)

    # remove a useless key which causes confusion
    if None in bijection.prime2x:
        bijection.prime2x.pop(None)

    return PrecisionRecallF1Output(
        f1=f1,
        precision=precision,
        recall=recall,
        bijection=bijection,
        n_corrects=TP,
        n_examples=len(all_groups.X_triples),
        n_predictions=len(all_groups.X_prime_triples),
        gold_triples=all_groups.X_triples,
        pred_triples=all_groups.X_prime_triples,
    )



import glob
import re
from contextlib import contextmanager
from multiprocessing.pool import Pool, ThreadPool
from operator import itemgetter
from pathlib import Path
from typing import Dict, Generic, TypeVar, Union, Callable, Any, List, Optional, KeysView

from loguru import logger
from tqdm.auto import tqdm


def str2bool(x):
    assert x in {"True", "False", "true", "false", "null"}
    if x == "<null>":
        return None
    return x.lower() == "true"


def nullable_str(x):
    if x == "<null>":
        return None
    return x


def str2int(x):
    if x == "<null>":
        return None
    return int(x)


def percentage(a, b):
    return "%.2f%% (%d/%d)" % (a * 100 / b, a, b)


def filter_duplication(lst: List[Any], key_fn: Callable[[Any], Any] = None):
    key_fn = key_fn or identity_func
    keys = set()
    new_lst = []
    for item in lst:
        k = key_fn(item)
        if k in keys:
            continue

        keys.add(k)
        new_lst.append(item)
    return new_lst


def identity_func(x):
    return x


def get_latest_version(file_pattern: Union[str, Path]) -> int:
    """Assuming the file pattern select list of files tagged with an integer version for every run, this
    function return the latest version number that you can use to name your next run.

    For example:
    1. If your pattern matches folders: version_1, version_5, version_6, this function will return 6.
    2. If your pattern does not match anything, return 0
    """
    files = [Path(file) for file in sorted(glob.glob(str(file_pattern)))]
    if len(files) == 0:
        return 0

    file = sorted(files)[-1]
    match = re.match("[^0-9]*(\d+)[^0-9]*", file.name)
    if match is None:
        raise Exception("Invalid naming")
    return int(match.group(1))


def get_incremental_path(path: Union[str, Path]) -> str:
    path = Path(str(path))
    if path.suffix == "":
        char = "_"
    else:
        char = "."

    pattern = path.parent / f"{path.stem}{char}*{path.suffix}"
    version = get_latest_version(pattern) + 1

    return str(path.parent / f"{path.stem}{char}{version:02d}{path.suffix}")


def get_latest_path(path: Union[str, Path]) -> Optional[str]:
    path = Path(str(path))
    pattern = path.parent / f"{path.stem}.*{path.suffix}"
    version = get_latest_version(pattern)
    if version == 0:
        return None
    return str(path.parent / f"{path.stem}.{version:02d}{path.suffix}")


def auto_wrap(word: str, max_char_per_line: int, delimiters: List[str] = None, camelcase_split: bool = True) -> str:
    """
    Treat this as optimization problem, where we trying to minimize the number of line break
    but also maximize the readability in each line, i.e: maximize the number of characters in each lines

    Using greedy search.
    :param word:
    :param max_char_per_line:
    :param delimiters:
    :return:
    """
    # split original word by the delimiters
    if delimiters is None:
        delimiters = [' ', ':', '_', '/']

    sublines: List[str] = [""]
    for i, c in enumerate(word):
        if c not in delimiters:
            sublines[-1] += c

            if camelcase_split and not c.isupper() and i + 1 < len(word) and word[i + 1].isupper():
                # camelcase_split
                sublines.append("")
        else:
            sublines[-1] += c
            sublines.append("")

    new_sublines: List[str] = [""]
    for line in sublines:
        if len(new_sublines[-1]) + len(line) <= max_char_per_line:
            new_sublines[-1] += line
        else:
            new_sublines.append(line)

    return "\n".join(new_sublines)


def flatten_dict(odict: dict, result: Optional[dict] = None, prefix: str = ""):
    if result is None:
        result = {}

    for k, v in odict.items():
        if isinstance(v, dict):
            flatten_dict(v, result, prefix=prefix + k + ".")
        else:
            result[prefix + k] = v
    return result


class ParallelMapFnWrapper:
    def __init__(self, fn):
        self.fn = fn

    def run(self, args):
        idx, r = args
        try:
            r = self.fn(r)
            return idx, r
        except:
            logger.error(f"[ParallelMap] Error while process item {idx}")
            raise


def parallel_map(fn, inputs, show_progress=False, progress_desc="", is_parallel=True, use_threadpool=False, n_processes: Optional[int]=None):
    if not is_parallel:
        iter = (fn(item) for item in inputs)
        if show_progress:
            iter = tqdm(iter, total=len(inputs), desc=progress_desc)
        return list(iter)

    if use_threadpool:
        with ThreadPool(processes=n_processes) as pool:
            iter = pool.imap_unordered(ParallelMapFnWrapper(fn).run, enumerate(inputs))
            if show_progress:
                iter = tqdm(iter, total=len(inputs), desc=progress_desc)
            results = list(iter)
            results.sort(key=itemgetter(0))
    else:
        with Pool(processes=n_processes) as pool:
            iter = pool.imap_unordered(ParallelMapFnWrapper(fn).run, enumerate(inputs))
            if show_progress:
                iter = tqdm(iter, total=len(inputs), desc=progress_desc)
            results = list(iter)
            results.sort(key=itemgetter(0))
    return [v for i, v in results]


@contextmanager
def print2file(file_path: Union[str, Path], mode="w", file_only: bool = False):
    """Yield a print function that can be both print to file or print to std"""
    Path(file_path).parent.mkdir(exist_ok=True, parents=True)
    origin_print = print
    with open(str(file_path), mode) as f:
        def print_fn(*args):
            if not file_only:
                origin_print(*args)
            origin_print(*args, file=f)

        try:
            yield print_fn
        finally:
            pass


class FakeTQDM:
    def update(self, *args):
        pass

    def close(self):
        pass


K = TypeVar('K')
V = TypeVar('V')
V2 = TypeVar('V2')


class DictProxy(Dict[K, V2]):
    """Dictionary proxy to access objects' property

    Args:
        odict: dictionary of object
        access: function to access property of an object
    """

    def __init__(self, odict: Dict[K, V], access: Callable[[V], V2]):
        self.odict = odict
        self.access = access

    def __iter__(self):
        return self.odict.__iter__()

    def __getitem__(self, item):
        return self.access(self.odict[item])

    def __contains__(self, item):
        return item in self.odict

    def keys(self) -> KeysView[K]:
        return self.odict.keys()

    def values(self):
        return (self.access(v) for v in self.odict.values())

    def items(self):
        return ((k, self.access(v)) for k, v in self.odict.items())



from typing import Dict, Callable, Union, TypeVar

from sm.misc.funcs import identity_func


K = TypeVar('K')
V = TypeVar('V')


class InMemStore(Dict[K, V]):
    def __init__(self, rdict: Dict[K, V], deserialize: Callable[[str], V] = None):
        self.rdict = rdict
        self.deserialize = deserialize or identity_func

    def __contains__(self, item):
        return item in self.rdict

    def __getitem__(self, item):
        return self.deserialize(self.rdict[item])

    def __setitem__(self, key: str, value: Union[str, bytes]):
        self.rdict[key] = value

    def values(self):
        return (self.deserialize(x) for x in self.rdict.values())

    def items(self):
        return ((k, self.deserialize(v)) for k, v in self.rdict.items())

    def keys(self):
        return self.rdict.keys()

    def __len__(self):
        return len(self.rdict)

    def cache_dict(self) -> 'CacheDictStore[K, V]':
        return CacheDictStore(self)

    def as_dict(self):
        return {k: self.deserialize(v) for k, v in self.rdict.items()}


class CacheDictStore(Dict[K, V]):
    def __init__(self, store: Dict[K, V]):
        self.store = store
        self.cache = {}

    def __contains__(self, item: str):
        return item in self.cache or item in self.store

    def __getitem__(self, item: str):
        if item not in self.cache:
            self.cache[item] = self.store[item]
        return self.cache[item]

    def __setitem__(self, key: str, value: Union[str, bytes]):
        raise Exception("NotSupportedFunction")

    def values(self):
        return self.store.values()

    def items(self):
        return self.store.items()

    def keys(self):
        return self.store.keys()

    def __len__(self):
        return len(self.store)


from operator import itemgetter
from typing import List, Dict, Tuple, Callable, Any, Optional

from loguru import logger

from sm_unk.config import HOME_DIR
from sm_unk.misc.graph import *

g1 = load_json(HOME_DIR / "graph_viz/graph.json")
g2 = load_json(HOME_DIR / "graph_viz/graph.json")

print(len(g1.nodes), len(g2.nodes))
print(len(g1.edges), len(g2.edges))

for nid, ndata in g1.nodes(data=True):
    assert nid in g2.nodes
for uid, vid, eid, edata in g1.edges(data=True, keys=True):
    assert (uid, vid, eid) in g2.edges, (uid, vid, eid)
    assert edata['data']['weight'] == g2.edges[(uid, vid, eid)]['data']['weight']


def maximum_spanning_tree_solver(origin_sm_g: nx.MultiDiGraph):
    sm_g = nx.MultiDiGraph()
    for uid, vid, eid, edata in origin_sm_g.edges(data=True, keys=True):
        edata['weight'] = edata['data']['weight']
    resp = nx.algorithms.tree.branchings.maximum_spanning_arborescence(origin_sm_g, attr='weight',
                                                                       preserve_attrs=True)
    count = 0
    for s, t, e, edata in resp.edges(data=True, keys=True):
        edge = edata['data']
        if s not in sm_g:
            sm_g.add_node(s, **origin_sm_g.nodes[s])
        if t not in sm_g:
            sm_g.add_node(t, **origin_sm_g.nodes[t])
        sm_g.add_edge(s, t, key=edge['predicate'], data=edge)
        count += edge['weight']
    print(count)


# @logger.catch
def solver2(g: nx.MultiDiGraph):
    new_g = {}
    roots = []
    for uid in g:
        new_g[uid] = {}
        if not uid.startswith("stmt:"):
            roots.append(uid)

        for vid in g[uid]:
            max_edge = max([(eid, edata['data']['weight']) for eid, edata in g[uid][vid].items()], key=itemgetter(1))
            new_g[uid][vid] = {
                "eid": max_edge[0],
                "weight": max_edge[1]
            }

    from sm_unk.misc.graph.edmonds import mst

    input_g = {
        uid: {vid: 1 / edata['weight'] for vid, edata in u.items()}
        for uid, u in new_g.items()
    }

    sols = []
    sol_weights = float('inf')
    for root in roots:
        resp = mst(root, input_g)
        resp_weight = sum(weight for uid, u in resp.items() for vid, weight in u.items())
        if resp_weight < sol_weights:
            if abs(resp_weight - sol_weights) > 1e-7:
                # not number error
                sols = []
                sol_weights = resp_weight
            sols.append(resp)

    for sol in sols:
        resp_g = nx.MultiDiGraph()
        for uid, u in sol.items():
            resp_g.add_node(uid, **g.nodes[uid])
            for vid, weight in u.items():
                if vid not in resp_g:
                    resp_g.add_node(vid, **g.nodes[vid])

                assert new_g[uid][vid]['weight'] == 1.0 / weight
                resp_g.add_edge(uid, vid, key=new_g[uid][vid]['eid'], **g.edges[uid, vid, new_g[uid][vid]['eid']])

        weight = sum(edata['data']['weight'] for uid, vid, eid, edata in resp_g.edges(data=True, keys=True))
        print(weight)
        return resp_g
    print(sol_weights)
    print(sols)


maximum_spanning_tree_solver(g1)
# maximum_spanning_tree_solver(g2)
g1 = solver2(g1)
# g2 = solver2(g2)
viz_graph(g1, lambda nid, n: {"label": n['data']['label']}, lambda eid, e: {"label": e['data']['predicate'] + f" #{e['data']['weight']}"}, HOME_DIR / "graph_viz", "g1")
# viz_graph(g2, lambda nid, n: {"label": n['data']['label']}, lambda eid, e: {"label": e['data']['predicate'] + f" #{e['data']['weight']}"}, HOME_DIR / "graph_viz", "g2")


import os
import shutil
import time
from pathlib import Path
from typing import Optional, Dict, Union

import pandas as pd

from sm.misc.deser import deserialize_csv, serialize_csv, serialize_json
from sm.misc.funcs import get_incremental_path
from sm.misc.timer import Timer


class ExpManager:
    def __init__(self, workdir: Union[Path, str], exp_name: str, exp_params: dict, integrate_neptune: bool = False):
        # directory containing the experiments
        self.workdir = Path(workdir)
        # name of the experiment: e.g., baseline, proposed-method-a, proposed-method-b
        self.exp_name = exp_name
        # parameters of the experiment
        self.exp_params = exp_params
        self.current_run_dir = None

        # whether we should integrating with neptune AI to track experiment run.
        self.integrate_neptune = integrate_neptune
        if integrate_neptune:
            import neptune
            self.neptune = neptune
            with Timer.get_instance().watch_and_report('Setup neptune take'):
                self.neptune.init(project_qualified_name=os.environ['NEPTUNE_PROJECT'],
                                  api_token=os.environ['NEPTUNE_API_TOKEN'])

        assert self.workdir.exists(), f"{self.workdir} does not exist"

    def create_run(self, run_id: Optional[str] = None):
        """Create new directory to store the run

        Parameters
        ----------
        run_id: optional run id (for reuse previous run)

        Returns
        -------
        the run directory
        """
        if run_id is not None:
            self.current_run_dir = self.workdir / run_id
        else:
            self.current_run_dir = Path(get_incremental_path(self.workdir / "v"))
        self.current_run_dir.mkdir(exist_ok=True, parents=True)
        serialize_json(self.exp_params, self.current_run_dir / "exp_params.json", indent=4)

        if self.integrate_neptune:
            with Timer.get_instance().watch_and_report('Create neptune experiment take'):
                neptune_exp_params = dict()
                neptune_exp_params.update(self.exp_params)
                neptune_exp_params['workdir'] = self.workdir
                neptune_exp_params['run_id'] = self.current_run_dir.name
                self.neptune.create_experiment(
                    name=self.exp_name, params=neptune_exp_params,
                    send_hardware_metrics=False, run_monitoring_thread=False)
        return self.current_run_dir

    def snapshot(self):
        assert self.current_run_dir is not None, "Need to create a run first."
        snapshot(self.current_run_dir)

    def report_metrics(self, metrics: Dict[str, Union[str, int, float]], filename: str = "metrics.csv"):
        if self.integrate_neptune:
            for k, v in metrics.items():
                if isinstance(v, (float, int)):
                    self.neptune.log_metric(k, v)
                else:
                    self.neptune.log_text(k, v)

        # log the data to a file
        pd.DataFrame([
            {"name": k, "value": v}
            for k, v in metrics.items()
        ]).to_csv(self.current_run_dir / filename, columns=["name", "value"])


class ExpResults:
    """Storing the experiment result in tsv/csv file.
    Old result is override by new result.
    Always keep a backup.
    """

    def __init__(self, outfile: Union[Path, str]):
        self.outfile = Path(outfile)
        self.records = {}
        self.headers = None
        assert self.outfile.name.endswith(".csv") or self.outfile.name.endswith(".tsv")
        if self.outfile.name.endswith(".csv"):
            self.delimiter = ","
        else:
            self.delimiter = "\t"

        if self.outfile.exists():
            records = deserialize_csv(self.outfile, self.delimiter)
            self.headers = records[0]
            for record in records[1:]:
                record = dict(zip(self.headers, record))
                self.records[record['id']] = record
            shutil.copyfile(str(self.outfile), str(self.outfile) + ".backup")

    def add(self, id: str, result: dict):
        """Add an result, identified by id, the result dict should not have id"""
        # copy record to keep the header in order, including new field called id
        new_record = {'id': id}
        for k, v in result.items():
            new_record[k] = v
        if self.headers is None:
            self.headers = list(new_record.keys())
            serialize_csv([self.headers], self.outfile, mode="a", delimiter=self.delimiter)

        self.records[id] = new_record
        serialize_csv([[new_record[k] for k in self.headers]], self.outfile, mode="a", delimiter=self.delimiter)

    def get_dataframe(self):
        return pd.DataFrame(self.records)


def snapshot(outdir: str, cwd=None):
    """Snapshot your code"""
    outdir = os.path.abspath(outdir)
    assert os.path.exists(outdir), f"Direction \"{outdir}\" does not exist"
    commit_file = os.path.join(outdir, "commit.txt")
    changes_file = os.path.join(outdir, "changes.patch")

    if cwd is None:
        cwd = os.path.dirname(os.path.abspath(__file__))

    print("Saving your changes", end="", flush="")
    start_time = time.time()
    # save current commit id
    commit_id = subprocess.check_output("git rev-parse HEAD", shell=True, cwd=cwd)
    commit_id = commit_id.decode().strip()
    with open(commit_file, "w") as f:
        f.write(commit_id)
    print(".", end="", flush=True)

    # save current change
    stash_id = subprocess.check_output("git stash create 'save your changes'", shell=True, cwd=cwd)
    stash_id = stash_id.decode().strip()

    if stash_id == "":
        # there is no change in tracked file
        print(" No changes in tracked files", end="", flush=True)
    else:
        subprocess.check_call(f"git stash store -m 'save your changes' {stash_id}", shell=True, cwd=cwd)
        print(".", end="", flush=True)
        subprocess.check_call(f"git stash show -p --binary > {changes_file}", shell=True, cwd=cwd)
        print(".", end="", flush=True)
        subprocess.check_call(f"git stash drop -q", shell=True, cwd=cwd)

    duration = time.time() - start_time
    print(f". DONE! (Taking {duration:.2f} seconds)", flush=True)


if __name__ == '__main__':
    import neptune

    neptune.init(project_qualified_name=os.environ['NEPTUNE_PROJECT'], api_token=os.environ['NEPTUNE_API_TOKEN'])
    neptune_exp_params = dict(test=True, name='PSL')
    neptune.create_experiment(name='Tester', params=neptune_exp_params)
    neptune.log_metric("noigi", 500)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sm',
 'sm.evaluation',
 'sm.inputs',
 'sm.misc',
 'sm.misc.big_dict',
 'sm.misc.graph',
 'sm.outputs']

package_data = \
{'': ['*'], 'sm': ['data/*']}

install_requires = \
['chardet>=4.0.0,<5.0.0',
 'ipython>=7.23.1,<8.0.0',
 'loguru>=0.5.3,<0.6.0',
 'matplotlib>=3.4.2,<4.0.0',
 'networkx>=2.5.1,<3.0.0',
 'orjson>=3.5.2,<4.0.0',
 'pandas>=1.2.4,<2.0.0',
 'pydot>=1.4.2,<2.0.0',
 'pyrsistent>=0.17.3,<0.18.0',
 'python-slugify>=5.0.2,<6.0.0',
 'redis>=3.5.3,<4.0.0',
 'rocksdb>=0.7.0,<0.8.0',
 'ruamel.yaml>=0.17.4,<0.18.0',
 'tqdm>=4.60.0,<5.0.0',
 'ujson>=4.0.2,<5.0.0']

setup_kwargs = {
    'name': 'sem-desc',
    'version': '0.1.13',
    'description': 'Package providing basic functionalities for the semantic modeling problem',
    'long_description': None,
    'author': 'Binh Vu',
    'author_email': 'binh@toan2.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
